"""
# Installation

`pip install youtube-up`

## Installing certificates

On your first run you may get an error which says `Was not able to load https://youtube.com. Have you installed the certificate at {cert_path} ?`.
If this happens you should follow the instructions at https://docs.mitmproxy.org/stable/concepts-certificates/#installing-the-mitmproxy-ca-certificate-manually
to install the certificate at the given path.

# Examples

## Upload a video
```python
from youtube_up import AllowCommentsEnum, Metadata, PrivacyEnum, YTUploaderSession

uploader = YTUploaderSession.from_cookies_txt("cookies/cookies.txt")
metadata = Metadata(
    title="Video title",
    description="Video description",
    privacy=PrivacyEnum.PUBLIC,
    made_for_kids=False,
    allow_comments_mode=AllowCommentsEnum.HOLD_ALL,
)
uploader.upload("video.webm", metadata)
```
Note that for Enum-type parameters we can either pass the Enum itself (as shown above),
or the Enum value, or the Enum key, as a string. For example, instead of writing

`allow_comments_mode=AllowCommentsEnum.HOLD_ALL`

we could instead write `allow_comments_mode="HOLD_ALL"`
or `allow_comments_mode="APPROVED_COMMENTS"`

## Upload multiple videos
```python
from youtube_up import Metadata, YTUploaderSession

uploader = YTUploaderSession.from_cookies_txt("cookies/cookies.txt")
metadata_1 = Metadata(
    title="Video 1",
)
metadata_2 = Metadata(
    title="Video 2",
)
uploader.upload("video1.webm", metadata_1)
uploader.upload("video2.webm", metadata_2)
```

## Upload to a new or existing playlist
```python
from youtube_up import Metadata, YTUploaderSession, Playlist

uploader = YTUploaderSession.from_cookies_txt("cookies/cookies.txt")
metadata = Metadata(
    title="Video 1",
    playlists=[
        Playlist(
            "Songs by me",
            description="Playlist that will only be created if "
            "no playlist exists with the title 'Songs by me'",
            create_if_title_doesnt_exist=True,
            create_if_title_exists=False,
        ),
        Playlist(
            "test playlist",
            description="Playlist that video will be added to "
            "only if it exists already. This description does "
            "nothing.",
            create_if_title_doesnt_exist=False,
            create_if_title_exists=False,
        ),
        Playlist(
            "Album",
            description="Playlist will be created even if there"
            " is already a playlist with the name 'Album'"
            create_if_title_doesnt_exist=True,
            create_if_title_exists=True,
        ),
    ],
)
uploader.upload("video.webm", metadata)
```

## CLI
youtube-up comes with a CLI app for uploading videos. For example, if we wanted to
create a public video with the title "Video title", we would execute the following command:
`youtube-up video video.webm --title="Video title" --cookies_file="cookies/cookies.txt" --privacy="PUBLIC"`

The app can also take a JSON file as input. For example, the following JSON file would upload
one video to a new or existing playlist called "Music" and one video which is set to premiere
on December 25th, 2023 at 5 PM (local time).

```json
[
    {
        "file": "song.webm",
        "metadata": {
            "title": "New song",
            "privacy": "PUBLIC",
            "playlists": [
                {
                    "title": "Music"
                }
            ]
        }
    },
    {
        "file": "premiere.webm",
        "metadata": {
            "title": "Special Announcement",
            "scheduled_upload": "2023-12-25T17:00:00",
            "premiere_countdown_duration": "ONE_MIN",
            "premiere_theme": "BRIGHT"
        }
    }
]
```

If we wanted the video to premiere at 5 PM GMT, would could have written "2023-12-25T17:00:00+00:00"
instead. We then run

`youtube-up json metadata.json --cookies_file="cookies/cookies.txt"`

to upload these videos.
"""

from .metadata import *
from .uploader import *

__all__ = uploader.__all__ + metadata.__all__



import argparse
import json

import tqdm

from .metadata import *
from .uploader import YTUploaderSession

try:
    from argparse import BooleanOptionalAction
except ImportError:
    from .polyfills import BooleanOptionalAction


def main():
    parser = argparse.ArgumentParser(
        prog="youtube-up",
        description="Upload videos to YouTube using the internal YouTube API",
    )
    subparsers = parser.add_subparsers(help="commands", dest="command", required=True)

    json_parser = subparsers.add_parser("json")
    json_parser.add_argument(
        "filename",
        help="JSON file specifying videos to upload. File should"
        "be an array of objects with 'file' and 'metadata' keys where 'file' "
        "is a path to a video file and 'metadata' is structured as the Metadata"
        "class",
    )
    json_parser.add_argument(
        "--cookies_file", help="Path to Netscape cookies.txt file", required=True
    )

    video_parser = subparsers.add_parser("video")
    video_parser.add_argument("filename", help="Video file to upload")
    video_parser.add_argument(
        "--cookies_file", help="Path to Netscape cookies.txt file", required=True
    )
    video_parser.add_argument("--title", help="Title. Max length 100", required=True)
    video_parser.add_argument(
        "--description", help="Description. Max length 5000", default=""
    )
    video_parser.add_argument(
        "--privacy", help="Privacy", type=PrivacyEnum, default=PrivacyEnum.PRIVATE
    )
    video_parser.add_argument(
        "--made_for_kids",
        help="Made for kids. If true comments will be disabled",
        action=BooleanOptionalAction,
        default=False,
    )
    video_parser.add_argument("--tags", nargs="+", help="List of tags", default=[])
    video_parser.add_argument(
        "--scheduled_upload",
        help="Date to make upload public, in ISO format (local time, unless timezone is specified)."
        " If set, video will be set to private until the date, unless video is a premiere in which "
        "case it will be set to public. Video will not be a premiere unless both "
        "premiere_countdown_duration and premiere_theme are set",
    )
    video_parser.add_argument(
        "--premiere_countdown_duration",
        help="Duration of premiere countdown in seconds",
        type=PremiereDurationEnum,
    )
    video_parser.add_argument(
        "--premiere_theme",
        help="Theme of premiere countdown",
        type=PremiereThemeEnum,
    )
    video_parser.add_argument(
        "--playlist_ids",
        nargs="+",
        help="List of existing playlist IDs to add video to",
    )
    video_parser.add_argument("--thumbnail", help="Path to thumbnail file to upload")
    video_parser.add_argument(
        "--publish_to_feed",
        help="Whether to notify subscribers",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--category",
        help="Category. Category-specific metadata is not supported yet",
        type=CategoryEnum,
    )
    video_parser.add_argument(
        "--auto_chapter",
        help="Whether to use automatic video chapters",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--auto_places",
        help="Whether to use automatic places",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--auto_concepts",
        help="Whether to use automatic concepts",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--has_product_placement",
        help="Whether video has product placement",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--show_product_placement_overlay",
        help="Whether to show product placement overlay",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--recorded_date",
        help="Day, month, and year that video was recorded, in ISO format",
    )
    video_parser.add_argument(
        "--restricted_to_over_18",
        help="Whether video is age restricted",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--audio_language",
        help="Language of audio. If uploading captions this must be set",
    )
    video_parser.add_argument(
        "--captions_file",
        help="Path to captions file (.srt) with language audio_language",
        type=CaptionsFile,
    )
    video_parser.add_argument(
        "--license",
        help="License",
        type=LicenseEnum,
    )
    video_parser.add_argument(
        "--allow_comments",
        help="Whether to allow comments",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--allow_comments_mode",
        help="Comment filtering mode",
        type=AllowCommentsEnum,
    )
    video_parser.add_argument(
        "--can_view_ratings",
        help="Whether video likes/dislikes can be seen",
        action=BooleanOptionalAction,
    )
    video_parser.add_argument(
        "--comments_sort_order",
        help="Default comment sort order",
        type=CommentsSortOrderEnum,
    )
    video_parser.add_argument(
        "--allow_embedding",
        help="Whether to allow embedding on 3rd party sites",
        action=BooleanOptionalAction,
    )

    args = parser.parse_args()

    uploader = YTUploaderSession.from_cookies_txt(args.cookies_file)

    if args.command == "json":
        with open(args.filename, "r") as f:
            data = json.load(f)
        with tqdm.tqdm(total=100 * len(data)) as pbar:
            for i, video in enumerate(data):

                def callback(step: str, prog: int):
                    pbar.n = 100 * i + prog
                    pbar.update()

                video_id = uploader.upload(
                    video["file"], Metadata.from_dict(video["metadata"]), callback
                )
                tqdm.tqdm.write(
                    f"Uploaded video: https://youtube.com/watch?v={video_id}"
                )
    else:
        args_dict = vars(args)
        args_dict.pop("cookies_file")
        args_dict.pop("command")
        video_file = args_dict.pop("filename")
        captions_file = args_dict.pop("captions_file")
        if captions_file is None:
            args_dict["captions_files"] = None
        else:
            args_dict["captions_files"] = [captions_file]
        metadata = Metadata.from_dict(args_dict)
        with tqdm.tqdm(total=100) as pbar:

            def callback(step: str, prog: int):
                pbar.n = prog
                pbar.update()

            video_id = uploader.upload(video_file, metadata, callback)
            tqdm.tqdm.write(f"Uploaded video: https://youtube.com/watch?v={video_id}")


if __name__ == "__main__":
    main()



import datetime
from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import config, dataclass_json

from .metadata import *


@dataclass_json
@dataclass(frozen=True)
class APIClient:
    clientName: int = 62
    clientVersion: str = "1.20231215.01.00"
    experimentsToken: str = ""
    gl: str = "US"
    hl: str = "en"
    utcOffsetMinutes: int = -300
    userInterfaceTheme: str = "USER_INTERFACE_THEME_DARK"
    screenWidthPoints: int = 1920
    screenHeightPoints: int = 529
    screenPixelDensity: int = 1
    screenDensityFloat: int = 1


@dataclass_json
@dataclass(frozen=True)
class APISessionInfo:
    token: str = ""


@dataclass_json
@dataclass(frozen=True)
class APIRequest:
    internalExperimentFlags: List = ()
    returnLogEntry: bool = True
    sessionInfo: APISessionInfo = APISessionInfo()


@dataclass_json
@dataclass(frozen=True)
class APIRoleType:
    channelRoleType: str = "CREATOR_CHANNEL_ROLE_TYPE_OWNER"


@dataclass_json
@dataclass(frozen=True)
class APIDelegationContext:
    externalChannelId: str
    roleType: APIRoleType = APIRoleType()


@dataclass_json
@dataclass(frozen=True)
class APIUser:
    delegationContext: APIDelegationContext
    onBehalfOfUser: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )


@dataclass_json
@dataclass(frozen=True)
class APIContext:
    client: APIClient
    request: APIRequest
    user: APIUser

    @classmethod
    def from_session_data(
        cls, channel_id: str, session_token: str, delegated_session_id: str
    ):
        return cls(
            APIClient(),
            APIRequest(sessionInfo=APISessionInfo(session_token)),
            APIUser(
                APIDelegationContext(channel_id), onBehalfOfUser=delegated_session_id
            ),
        )


@dataclass_json
@dataclass(frozen=True)
class APIID:
    id: str


@dataclass_json
@dataclass(frozen=True)
class APIScottyResourceID:
    scottyResourceId: APIID


@dataclass_json
@dataclass(frozen=True)
class APIMetadataTitle:
    newTitle: str


@dataclass_json
@dataclass(frozen=True)
class APIMetadataDescription:
    newDescription: str
    shouldSegment: bool = True


@dataclass_json
@dataclass(frozen=True)
class APIMetadataPrivacy:
    newPrivacy: str


@dataclass_json
@dataclass(frozen=True)
class APIMetadataDraftState:
    isDraft: bool


@dataclass_json
@dataclass(frozen=True)
class APIMetadataTags:
    newTags: List[str]


class UpdateMetadataBase:
    @classmethod
    def from_metadata_args(cls, *args):
        if all(a is None for a in args):
            return None
        else:
            return cls(*args)


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataMadeForKids(UpdateMetadataBase):
    newMfk: str
    operation: str = "MDE_MADE_FOR_KIDS_UPDATE_OPERATION_SET"


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataPrivacy(UpdateMetadataBase):
    newPrivacy: str


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataCategory(UpdateMetadataBase):
    newCategoryId: int


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataCommentOptions(UpdateMetadataBase):
    newAllowComments: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    newAllowCommentsMode: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    newCanViewRatings: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    newDefaultSortOrder: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataDistributionOptions(UpdateMetadataBase):
    newAllowEmbedding: bool


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataLicense(UpdateMetadataBase):
    newLicenseId: str


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataPublishingOptions(UpdateMetadataBase):
    newPostToFeed: bool


@dataclass_json
@dataclass(frozen=True)
class APIDate:
    day: int
    month: int
    year: int

    @classmethod
    def from_date(cls, date: Optional[datetime.datetime]):
        if date is None:
            return None
        else:
            return cls(date.day, date.month, date.year)


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataRecordedDate(UpdateMetadataBase):
    newRecordedDate: APIDate
    operation: str = "MDE_RECORDED_DATE_UPDATE_OPERATION_SET"


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataAudioLanguage(UpdateMetadataBase):
    newAudioLanguage: str


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataAutoChapter(UpdateMetadataBase):
    creatorOptOut: bool


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataAutoPlaces(UpdateMetadataBase):
    creatorOptOut: bool


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataAutoLearningConcepts(UpdateMetadataBase):
    autoConceptsCreatorOptOut: bool


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataProductPlacement(UpdateMetadataBase):
    newHasPaidProductPlacement: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    newShowPaidProductPlacementOverlay: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataRacy(UpdateMetadataBase):
    newRacy: str
    operation: str = "MDE_RACY_UPDATE_OPERATION_SET"


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataRemoveDraftState(UpdateMetadataBase):
    operation: str = "MDE_DRAFT_STATE_UPDATE_OPERATION_REMOVE_DRAFT_STATE"


@dataclass_json
@dataclass(frozen=True)
class APIImage(UpdateMetadataBase):
    encryptedScottyResourceId: str
    format: str
    name: str = "CUSTOM_THUMBNAIL_IMAGE_NAME_DEFAULT"


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataThumbnail(UpdateMetadataBase):
    image: APIImage
    operation: str = "UPLOAD_CUSTOM_THUMBNAIL"


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataPlaylists(UpdateMetadataBase):
    addToPlaylistIds: List[str]
    deleteFromPlaylistIds: List[str] = ()


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataSchedule(UpdateMetadataBase):
    timeSec: str
    privacy: str = "PUBLIC"

    @classmethod
    def from_date(cls, date: Optional[datetime.datetime]):
        if date is None:
            return None
        else:
            return cls(str(round(date.timestamp())))


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataScheduledPublishing(UpdateMetadataBase):
    set: APIUpdateMetadataSchedule


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataPremiere(UpdateMetadataBase):
    scheduledStartTimeSec: str
    operation: str = "MDE_PREMIERE_UPDATE_OPERATION_SCHEDULE"

    @classmethod
    def from_date(cls, date: Optional[datetime.datetime]):
        if date is None:
            return None
        else:
            return cls(str(round(date.timestamp())))


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataCountdown(UpdateMetadataBase):
    seconds: str


@dataclass_json
@dataclass(frozen=True)
class APIUpdateMetadataPremiereIntro(UpdateMetadataBase):
    countdownDuration: APIUpdateMetadataCountdown
    theme: str
    operation: str = "MDE_PREMIERE_INTRO_UPDATE_OPERATION_SET"


@dataclass_json
@dataclass(frozen=True)
class APIInitialMetadata:
    title: APIMetadataTitle
    description: APIMetadataDescription
    privacy: APIMetadataPrivacy
    draftState: APIMetadataDraftState
    tags: APIMetadataTags

    @classmethod
    def from_metadata(cls, metadata: Metadata):
        return cls(
            APIMetadataTitle(metadata.title),
            APIMetadataDescription(metadata.description),
            APIMetadataPrivacy(PrivacyEnum.PRIVATE),
            APIMetadataDraftState(True),
            APIMetadataTags(metadata.tags),
        )


@dataclass_json
@dataclass(frozen=True)
class APIRequestCreateVideo:
    channelId: str
    context: APIContext
    delegationContext: APIDelegationContext
    frontendUploadId: str
    initialMetadata: APIInitialMetadata
    presumedShort: bool
    resourceId: APIScottyResourceID

    @classmethod
    def from_session_data(
        cls,
        channel_id: str,
        session_token: str,
        delegated_session_id: Optional[str],
        front_end_upload_id: str,
        metadata: Metadata,
        scotty_resource_id: str,
    ):
        return cls(
            channel_id,
            APIContext.from_session_data(
                channel_id, session_token, delegated_session_id
            ),
            APIDelegationContext(channel_id),
            front_end_upload_id,
            APIInitialMetadata.from_metadata(metadata),
            False,  # apparently does nothing; will be a short if video meets short requirements regardless
            APIScottyResourceID(APIID(scotty_resource_id)),
        )


@dataclass_json
@dataclass(frozen=True)
class APIRequestListPlaylistsMask:
    playlistId: bool = True
    title: bool = True


@dataclass_json
@dataclass(frozen=True)
class APIRequestListPlaylists:
    channelId: str
    context: APIContext
    delegationContext: APIDelegationContext
    mask: APIRequestListPlaylistsMask = APIRequestListPlaylistsMask()
    memberVideoIds: List[str] = ()
    pageSize: int = 500

    @classmethod
    def from_session_data(
        cls, channel_id: str, session_token: str, delegated_session_id: Optional[str]
    ):
        return cls(
            channel_id,
            APIContext.from_session_data(
                channel_id, session_token, delegated_session_id
            ),
            APIDelegationContext(channel_id),
        )


@dataclass_json
@dataclass(frozen=True)
class APICaptionsFile:
    dataUri: str
    fileName: str


@dataclass_json
@dataclass(frozen=True)
class APICaptionsTrackData:
    lang: str
    kind: str = ""
    name: str = ""


@dataclass_json
@dataclass(frozen=True)
class APIOperationUpdateCaptions:
    captionsFile: APICaptionsFile
    ttsTrackId: APICaptionsTrackData
    contentUpdateTime: str
    isContentEdited: bool = False
    userIntent: str = "USER_INTENT_EDIT_LATEST_DRAFT"
    vote: str = "VOTE_PUBLISH"


@dataclass_json
@dataclass(frozen=True)
class APIRequestUpdateCaptions:
    channelId: str
    context: APIContext
    operations: List[APIOperationUpdateCaptions]
    videoId: str

    @classmethod
    def from_session_data(
        cls,
        channel_id: str,
        session_token: str,
        delegated_session_id: Optional[str],
        video_id: str,
        caption_file: str,
        caption_file_base64: str,
        caption_language: str,
        nanosecond_timestamp: str,
    ):
        return cls(
            channel_id,
            APIContext.from_session_data(
                channel_id, session_token, delegated_session_id
            ),
            APIOperationUpdateCaptions(
                APICaptionsFile(caption_file_base64, caption_file),
                APICaptionsTrackData(caption_language),
                nanosecond_timestamp,
            ),
            video_id,
        )


@dataclass_json
@dataclass(frozen=True)
class APIPlaylistCourseMetadata:
    isCourse: bool = False


@dataclass_json
@dataclass(frozen=True)
class APIPlaylistPodcastMetadata:
    isPodcast: bool = False


@dataclass_json
@dataclass(frozen=True)
class APIRequestCreatePlaylist:
    context: APIContext
    delegationContext: APIDelegationContext
    title: str
    description: str
    privacyStatus: str
    courseMetadata: APIPlaylistCourseMetadata = APIPlaylistCourseMetadata()
    podcastMetadata: APIPlaylistPodcastMetadata = APIPlaylistPodcastMetadata()

    @classmethod
    def from_session_data(
        cls,
        channel_id: str,
        session_token: str,
        delegated_session_id: Optional[str],
        playlist: Playlist,
    ):
        return cls(
            APIContext.from_session_data(
                channel_id, session_token, delegated_session_id
            ),
            APIDelegationContext(channel_id),
            playlist.title,
            playlist.description,
            playlist.privacy,
        )


RacyDict = {
    True: "MDE_RACY_TYPE_RESTRICTED",
    False: "MDE_RACY_TYPE_NOT_RESTRICTED",
    None: None,
}
MFKDict = {True: "MDE_MADE_FOR_KIDS_TYPE_MFK", False: "MDE_MADE_FOR_KIDS_TYPE_NOT_MFK"}


@dataclass_json
@dataclass(frozen=True)
class APIRequestUpdateMetadata:
    context: APIContext
    delegationContext: APIDelegationContext
    encryptedVideoId: str
    # mandatory updates
    madeForKids: APIUpdateMetadataMadeForKids
    draftState: APIUpdateMetadataRemoveDraftState
    privacyState: APIUpdateMetadataPrivacy
    # optional updates
    autoChapter: Optional[APIUpdateMetadataAutoChapter] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    autoPlaces: Optional[APIUpdateMetadataAutoPlaces] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    learningConcepts: Optional[APIUpdateMetadataAutoLearningConcepts] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    productPlacement: Optional[APIUpdateMetadataProductPlacement] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    racy: Optional[APIUpdateMetadataRacy] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    audioLanguage: Optional[APIUpdateMetadataAudioLanguage] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    recordedDate: Optional[APIUpdateMetadataRecordedDate] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    category: Optional[APIUpdateMetadataCategory] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    commentOptions: Optional[APIUpdateMetadataCommentOptions] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    distributionOptions: Optional[APIUpdateMetadataDistributionOptions] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    license: Optional[APIUpdateMetadataLicense] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    publishingOptions: Optional[APIUpdateMetadataPublishingOptions] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    videoStill: Optional[APIUpdateMetadataThumbnail] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    addToPlaylist: Optional[APIUpdateMetadataPlaylists] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    scheduledPublishing: Optional[APIUpdateMetadataScheduledPublishing] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    premiere: Optional[APIUpdateMetadataPremiere] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    premiereIntro: Optional[APIUpdateMetadataPremiereIntro] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_session_data(
        cls,
        channel_id: str,
        session_token: str,
        delegated_session_id: Optional[str],
        encrypted_video_id: str,
        metadata: Metadata,
        thumbnail_scotty_id: Optional[str] = None,
        thumbnail_format: Optional[str] = None,
    ):
        if (
            metadata.premiere_countdown_duration is not None
            and metadata.premiere_theme is not None
        ):
            # premiere, not scheduled upload
            premier_upload_time = metadata.scheduled_upload
            scheduled_upload_time = None
            metadata.privacy = PrivacyEnum.PUBLIC
        elif metadata.scheduled_upload is not None:
            # scheduled upload
            premier_upload_time = None
            scheduled_upload_time = metadata.scheduled_upload
            metadata.privacy = PrivacyEnum.PRIVATE
        else:
            # neither scheduled upload nor premiere
            premier_upload_time = None
            scheduled_upload_time = None

        return cls(
            APIContext.from_session_data(
                channel_id, session_token, delegated_session_id
            ),
            APIDelegationContext(channel_id),
            encrypted_video_id,
            APIUpdateMetadataMadeForKids(MFKDict[metadata.made_for_kids]),
            APIUpdateMetadataRemoveDraftState(),
            APIUpdateMetadataPrivacy(metadata.privacy),
            APIUpdateMetadataAutoChapter.from_metadata_args(metadata.auto_chapter),
            APIUpdateMetadataAutoPlaces.from_metadata_args(metadata.auto_places),
            APIUpdateMetadataAutoLearningConcepts.from_metadata_args(
                metadata.auto_concepts
            ),
            APIUpdateMetadataProductPlacement.from_metadata_args(
                metadata.has_product_placement, metadata.show_product_placement_overlay
            ),
            APIUpdateMetadataRacy.from_metadata_args(
                RacyDict[metadata.restricted_to_over_18]
            ),
            APIUpdateMetadataAudioLanguage.from_metadata_args(metadata.audio_language),
            APIUpdateMetadataRecordedDate.from_metadata_args(
                APIDate.from_date(metadata.recorded_date)
            ),
            APIUpdateMetadataCategory.from_metadata_args(metadata.category),
            APIUpdateMetadataCommentOptions.from_metadata_args(
                metadata.allow_comments,
                metadata.allow_comments_mode,
                metadata.can_view_ratings,
                metadata.comments_sort_order,
            ),
            APIUpdateMetadataDistributionOptions.from_metadata_args(
                metadata.allow_embedding
            ),
            APIUpdateMetadataLicense.from_metadata_args(metadata.license),
            APIUpdateMetadataPublishingOptions.from_metadata_args(
                metadata.publish_to_feed
            ),
            APIUpdateMetadataThumbnail.from_metadata_args(
                APIImage.from_metadata_args(thumbnail_scotty_id, thumbnail_format)
            ),
            APIUpdateMetadataPlaylists.from_metadata_args(metadata.playlist_ids),
            APIUpdateMetadataScheduledPublishing.from_metadata_args(
                APIUpdateMetadataSchedule.from_date(scheduled_upload_time)
            ),
            APIUpdateMetadataPremiere.from_date(premier_upload_time),
            APIUpdateMetadataPremiereIntro.from_metadata_args(
                APIUpdateMetadataCountdown.from_metadata_args(
                    metadata.premiere_countdown_duration
                ),
                metadata.premiere_theme,
            ),
        )



import base64
import copy
import json
import math
import os
import re
import time
import uuid
from hashlib import sha1
from http.cookiejar import Cookie, FileCookieJar
from typing import Callable, Dict

import requests
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.firefox.service import Service as FirefoxService
from seleniumwire2 import webdriver
from seleniumwire2.utils import decode
from tqdm.utils import CallbackIOWrapper

from .metadata import *
from .metadata import CaptionsFile
from .schema import *

from http.cookiejar import MozillaCookieJar


class YTUploaderException(Exception):
    """YouTube uploader exception"""


@dataclass
class YTUploaderVideoData:
    authuser: str = None
    channel_id: str = None
    innertube_api_key: str = None
    delegated_session_id: str = None
    front_end_upload_id: str = None
    encrypted_video_id: str = None
    thumbnail_scotty_id: str = None
    thumbnail_format: str = None


class YTUploaderSession:
    """
    Class for uploading YouTube videos to a single channel
    """

    _delegated_session_id_regex = re.compile(r'"DELEGATED_SESSION_ID":"([^"]*)"')
    _innertube_api_key_regex = re.compile(r'"INNERTUBE_API_KEY":"([^"]*)"')
    _session_index_regex = re.compile(r'"SESSION_INDEX":"([^"]*)"')
    _channel_id_regex = re.compile(r"https://studio.youtube.com/channel/([^/]*)/*")
    _progress_steps = {
        "start": 0,
        "get_session_data": 10,
        "get_upload_url": 20,
        "upload_video": 70,
        "get_session_token": 80,
        "create_video": 90,
        "upload_thumbnail": 95,
        "finish": 100,
    }
    _cookie_whitelist = {
        "LOGIN_INFO",
        "__Secure-1PSID",
        "__Secure-3PSID",
        "__Secure-1PAPISID",
        "__Secure-3PAPISID",
        "__Secure-1PSIDTS",
        "__Secure-3PSIDTS",
        "SAPISID",
    }

    _session_token: str
    _cookies: FileCookieJar
    _session: requests.Session

    def __init__(
        self,
        cookie_jar: FileCookieJar,
        webdriver_path: Optional[str] = None,
        selenium_timeout: float = 60,
    ):
        """Create YTUploaderSession from generic FileCookieJar

        Args:
            cookie_jar (FileCookieJar): FileCookieJar. Must have save(), load(),
                and set_cookie(http.cookiejar.Cookie) methods
            webdriver_path (str, optional): Optional path to geckodriver or chromedriver executable
            selenium_timeout (float, optional): Timeout to wait for grst request. Defaults to 60 seconds
        """
        self._session_token = ""
        self._webdriver_path = webdriver_path
        self._selenium_timeout = selenium_timeout

        # load cookies and init session
        self._cookies = cookie_jar
        self._cookies.load(ignore_discard=True, ignore_expires=True)
        self._session = requests.Session()
        for cookie in self._cookies:
            if cookie.name == "SESSION_TOKEN":
                self._session_token = cookie.value
            elif cookie.name in self._cookie_whitelist:
                self._session.cookies.set_cookie(copy.copy(cookie))
        self._session.headers = {
            "Authorization": f"SAPISIDHASH {self._generateSAPISIDHASH(self._session.cookies['SAPISID'])}",
            "x-origin": "https://studio.youtube.com",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/119.0",
        }

    @classmethod
    def from_cookies_txt(
        cls,
        cookies_txt_path: str,
        webdriver_path: Optional[str] = None,
        selenium_timeout: float = 60,
    ):
        """Create YTUploaderSession from cookies.txt file

        Args:
            cookies_txt_path (str): Path to Netscape cookies format file
            webdriver_path (str, optional): Optional path to geckodriver or chromedriver executable
            selenium_timeout (float, optional): Timeout to wait for grst request. Defaults to 60 seconds
        """
        cj = MozillaCookieJar(cookies_txt_path)
        return cls(cj, webdriver_path, selenium_timeout)

    def upload(
        self,
        file_path: str,
        metadata: Metadata,
        progress_callback: Callable[[str, float], None] = lambda step, percent: None,
    ) -> str:
        """Upload a video

        Args:
            file_path (str): Path to video file
            metadata (Metadata): Metadata of video to set when uploaded
            progress_callback (Callable[[str, float], None], optional): Optional progress callback.
                Callback receives what step uploader is on and what the total percentage of the upload
                progress is (defined by YTUploaderSession._progress_steps).

        Returns:
            str: ID of video uploaded
        """
        try:
            metadata.validate()
        except ValueError as ex:
            raise YTUploaderException(f"Validation error: {ex}") from ex
        progress_callback("start", self._progress_steps["start"])
        data = YTUploaderVideoData()
        self._get_session_data(data)
        progress_callback("get_session_data", self._progress_steps["get_session_data"])
        url = self._get_video_upload_url(data)
        progress_callback("get_upload_url", self._progress_steps["get_upload_url"])
        scotty_resource_id = self._upload_file(
            url, file_path, progress_callback, "get_upload_url", "upload_video"
        )
        progress_callback("upload_video", self._progress_steps["upload_video"])
        data.encrypted_video_id = self._create_video(scotty_resource_id, metadata, data)
        if data.encrypted_video_id is None:
            # could be bad session token, try to get new one
            self._get_session_token()
            progress_callback(
                "get_session_token", self._progress_steps["get_session_token"]
            )
            data.encrypted_video_id = self._create_video(
                scotty_resource_id, metadata, data
            )
            if data.encrypted_video_id is None:
                raise YTUploaderException("Could not create video")
        progress_callback("create_video", self._progress_steps["create_video"])

        # set thumbnail
        if metadata.thumbnail is not None:
            url = self._get_upload_url_thumbnail(data)
            data.thumbnail_scotty_id = self._upload_file(
                url,
                metadata.thumbnail,
                progress_callback,
                "create_video",
                "upload_thumbnail",
            )
            data.thumbnail_format = self._get_thumbnail_format(metadata.thumbnail)

        # playlists
        if metadata.playlists:
            playlists = self._get_creator_playlists(data)
            if metadata.playlist_ids is None:
                metadata.playlist_ids = []
            for playlist in metadata.playlists:
                exists = playlist.title in playlists
                if (playlist.create_if_title_exists and exists) or (
                    playlist.create_if_title_doesnt_exist and not exists
                ):
                    playlist_id = self._create_playlist(playlist, data)
                    metadata.playlist_ids.append(playlist_id)
                elif exists:
                    metadata.playlist_ids.append(playlists[playlist.title])
        # captions
        if metadata.captions_files:
            for caption_file in metadata.captions_files:
                if caption_file.language is None:
                    caption_file.language = metadata.audio_language
                self._update_captions(caption_file, data)

        self._update_metadata(metadata, data)
        # save cookies
        for cookie in self._session.cookies:
            self._cookies.set_cookie(cookie)
        self._cookies.save()
        progress_callback("finish", self._progress_steps["finish"])
        return data.encrypted_video_id

    def has_valid_cookies(self) -> bool:
        """Check if cookies are valid

        Returns:
            bool: True if we are able to log in to YouTube with the given cookies
        """
        r = self._session.get("https://youtube.com/upload")

        return "studio.youtube.com/channel" in r.url

    def _get_thumbnail_format(self, filename: str) -> ThumbnailFormatEnum:
        ext = filename.split(".")[-1]
        if ext in ("jpg", "jpeg", "jfif", "pjpeg", "pjp"):
            return ThumbnailFormatEnum.JPG
        if ext in ("png",):
            return ThumbnailFormatEnum.PNG
        raise YTUploaderException(
            f"Unknown format for thumbnail with extension '{ext}'. Only JPEG and PNG allowed"
        )

    def _get_session_token(self):
        try:
            # try firefox
            options = webdriver.FirefoxOptions()
            options.add_argument("--headless")
            if self._webdriver_path:
                service = FirefoxService(self._webdriver_path)
                driver = webdriver.Firefox(options=options, service=service)
            else:
                driver = webdriver.Firefox(options=options)
        except Exception:
            try:
                # try chrome
                options = webdriver.ChromeOptions()
                options.add_argument("--headless=new")
                if self._webdriver_path:
                    service = ChromeService(self._webdriver_path)
                    driver = webdriver.Chrome(options=options, service=service)
                else:
                    driver = webdriver.Chrome(options=options)
            except Exception:
                raise YTUploaderException(
                    "Could not launch Firefox or Chrome. Make sure geckodriver or chromedriver is installed"
                )

        driver.set_page_load_timeout(self._selenium_timeout)

        try:
            driver.get("https://youtube.com")
        except Exception:
            cert_path = os.path.join(
                driver.backend.storage.home_dir, "mitmproxy-ca-cert.cer"
            )
            raise YTUploaderException(
                f"Was not able to load https://youtube.com. Have you installed the certificate at {cert_path} ? See https://docs.mitmproxy.org/stable/concepts-certificates/#installing-the-mitmproxy-ca-certificate-manually"
            )

        for cookie in self._cookies:
            if cookie.name in self._cookie_whitelist:
                driver.add_cookie(cookie.__dict__)

        driver.get("https://youtube.com/upload")

        if "studio.youtube.com/channel" not in driver.current_url:
            driver.quit()
            raise YTUploaderException(
                "Could not log in to YouTube account. Try getting new cookies"
            )

        r = driver.wait_for_request(
            "studio.youtube.com/youtubei/v1/ars/grst", timeout=self._selenium_timeout
        )
        response = r.response
        r_json = json.loads(
            decode(response.body, response.headers.get("Content-Encoding"))
        )
        self._session_token = r_json["sessionToken"]
        self._cookies.set_cookie(
            Cookie(
                None,
                "SESSION_TOKEN",
                self._session_token,
                None,
                False,
                "",
                False,
                False,
                "",
                False,
                False,
                None,
                False,
                None,
                None,
                {},
            )
        )
        self._cookies.save()
        driver.quit()

    @staticmethod
    def _generateUUID() -> str:
        return str(uuid.uuid4()).upper()

    @staticmethod
    def _generateSAPISIDHASH(SAPISID) -> str:
        timestamp = math.floor(time.time())
        msg = f"{timestamp} {SAPISID} {'https://studio.youtube.com'}"
        hash = sha1(msg.encode("utf-8")).hexdigest()
        return f"{timestamp}_{hash}"

    def _get_session_data(self, data: YTUploaderVideoData):
        r = self._session.get("https://youtube.com/upload")

        if "studio.youtube.com/channel" not in r.url:
            raise YTUploaderException(
                "Could not log in to YouTube account. Try getting new cookies"
            )

        data.channel_id = self._channel_id_regex.match(r.url).group(1)
        data.innertube_api_key = self._innertube_api_key_regex.search(r.text).group(1)
        m = self._delegated_session_id_regex.search(r.text)
        data.delegated_session_id = m and m.group(1)
        data.authuser = self._session_index_regex.search(r.text).group(1)
        self._session.headers["X-Goog-AuthUser"] = data.authuser

    def _get_upload_url(self, api_url: str, authuser: str, data: dict) -> str:
        params = {"authuser": authuser}
        headers = {
            "x-goog-upload-command": "start",
            "x-goog-upload-protocol": "resumable",
        }
        r = self._session.post(
            api_url,
            headers=headers,
            params=params,
            json=data,
        )
        r.raise_for_status()
        upload_url = r.headers["x-goog-upload-url"]
        return upload_url

    def _get_video_upload_url(self, data: YTUploaderVideoData) -> str:
        data.front_end_upload_id = f"innertube_studio:{self._generateUUID()}:0"
        return self._get_upload_url(
            "https://upload.youtube.com/upload/studio",
            data.authuser,
            {"frontendUploadId": data.front_end_upload_id},
        )

    def _get_upload_url_thumbnail(self, data: YTUploaderVideoData) -> str:
        return self._get_upload_url(
            "https://upload.youtube.com/upload/studiothumbnail", data.authuser, {}
        )

    def _get_creator_playlists(self, data: YTUploaderVideoData) -> Dict[str, str]:
        params = {"key": data.innertube_api_key, "alt": "json"}
        data = APIRequestListPlaylists.from_session_data(
            data.channel_id, self._session_token, data.delegated_session_id
        ).to_dict()
        r = self._session.post(
            "https://studio.youtube.com/youtubei/v1/creator/list_creator_playlists",
            params=params,
            json=data,
        )
        r.raise_for_status()
        return {
            playlist["title"]: playlist["playlistId"]
            for playlist in r.json().get("playlists", [])
        }

    def _create_playlist(
        self,
        playlist: Playlist,
        data: YTUploaderVideoData,
    ) -> str:
        params = {"key": data.innertube_api_key, "alt": "json"}
        data = APIRequestCreatePlaylist.from_session_data(
            data.channel_id, self._session_token, data.delegated_session_id, playlist
        ).to_dict()
        r = self._session.post(
            "https://studio.youtube.com/youtubei/v1/playlist/create",
            params=params,
            json=data,
        )
        r.raise_for_status()
        return r.json()["playlistId"]

    def _update_captions(
        self,
        caption_file: CaptionsFile,
        data: YTUploaderVideoData,
    ):
        params = {"key": data.innertube_api_key, "alt": "json"}
        with open(caption_file.path, "rb") as f:
            captions_b64 = "data:application/octet-stream;base64," + base64.b64encode(
                f.read()
            ).decode("utf-8")
        timestamp = str(time.time_ns())
        data = APIRequestUpdateCaptions.from_session_data(
            data.channel_id,
            self._session_token,
            data.delegated_session_id,
            data.encrypted_video_id,
            caption_file.path,
            captions_b64,
            caption_file.language,
            timestamp,
        ).to_dict()
        r = self._session.post(
            "https://studio.youtube.com/youtubei/v1/globalization/update_captions",
            params=params,
            json=data,
        )
        r.raise_for_status()

    def _upload_file(
        self,
        upload_url: str,
        file_path: str,
        progress_callback: Callable[[str, float], None],
        prev_progress_step: str,
        cur_progress_step: str,
    ):
        headers = {
            "x-goog-upload-command": "upload, finalize",
            "x-goog-upload-offset": "0",
        }

        with open(file_path, "rb") as f:
            f.seek(0, os.SEEK_END)
            size = f.tell()
            f.seek(0)
            bytes_sent = 0

            def upload_callback(bytes: int):
                nonlocal bytes_sent
                bytes_sent += bytes
                start_prog = self._progress_steps[prev_progress_step]
                end_prog = self._progress_steps[cur_progress_step]
                cur_prog = start_prog + (end_prog - start_prog) * (bytes_sent / size)
                cur_prog = round(cur_prog, 1)
                progress_callback(cur_progress_step, cur_prog)

            wrapped_file = CallbackIOWrapper(upload_callback, f)
            r = self._session.post(upload_url, headers=headers, data=wrapped_file)

        r.raise_for_status()
        return r.json()["scottyResourceId"]

    def _create_video(
        self, scotty_resource_id: str, metadata: Metadata, data: YTUploaderVideoData
    ) -> Optional[str]:
        params = {"key": data.innertube_api_key, "alt": "json"}
        data = APIRequestCreateVideo.from_session_data(
            data.channel_id,
            self._session_token,
            data.delegated_session_id,
            data.front_end_upload_id,
            metadata,
            scotty_resource_id,
        ).to_dict()
        r = self._session.post(
            "https://studio.youtube.com/youtubei/v1/upload/createvideo",
            params=params,
            json=data,
        )
        r.raise_for_status()
        r = r.json()
        return r.get("videoId")

    def _update_metadata(self, metadata: Metadata, data: YTUploaderVideoData):
        params = {"key": data.innertube_api_key, "alt": "json"}
        data = APIRequestUpdateMetadata.from_session_data(
            data.channel_id,
            self._session_token,
            data.delegated_session_id,
            data.encrypted_video_id,
            metadata,
            data.thumbnail_scotty_id,
            data.thumbnail_format,
        ).to_dict()
        r = self._session.post(
            "https://studio.youtube.com/youtubei/v1/video_manager/metadata_update",
            params=params,
            json=data,
        )
        r.raise_for_status()


__all__ = ["YTUploaderSession", "YTUploaderException"]



import datetime
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

import marshmallow.fields as mm_field
from dataclasses_json import config, dataclass_json


class PrivacyEnum(str, Enum):
    PRIVATE = "PRIVATE"
    UNLISTED = "UNLISTED"
    PUBLIC = "PUBLIC"


class CategoryEnum(int, Enum):
    FILM_ANIMATION = 1
    AUTOS_VEHICLES = 2
    MUSIC = 10
    PETS_ANIMALS = 15
    SPORTS = 17
    TRAVEL_EVENTS = 19
    GAMING = 20
    PEOPLE_BLOGS = 22
    COMEDY = 23
    ENTERTAINMENT = 24
    NEWS_POLITICS = 25
    HOWTO_STYLE = 26
    EDUCATION = 27
    SCIENCE_TECH = 28
    NONPROFITS_ACTIVISM = 29


class LanguageEnum(str, Enum):
    NOT_APPLICABLE = "zxx"
    ABKHAZIAN = "ab"
    AFAR = "aa"
    AFRIKAANS = "af"
    AKAN = "ak"
    AKKADIAN = "akk"
    ALBANIAN = "sq"
    AMERICAN_SIGN_LANGUAGE = "ase"
    AMHARIC = "am"
    ARABIC = "ar"
    ARAMAIC = "arc"
    ARMENIAN = "hy"
    ASSAMESE = "as"
    AYMARA = "ay"
    AZERBAIJANI = "az"
    BAMBARA = "bm"
    BANGLA = "bn"
    BANGLA_INDIA = "bn-IN"
    BASHKIR = "ba"
    BASQUE = "eu"
    BELARUSIAN = "be"
    BHOJPURI = "bho"
    BISLAMA = "bi"
    BODO = "brx"
    BOSNIAN = "bs"
    BRETON = "br"
    BULGARIAN = "bg"
    BURMESE = "my"
    CANTONESE = "yue"
    CANTONESE_HONG_KONG = "yue-HK"
    CATALAN = "ca"
    CHEROKEE = "chr"
    CHINESE = "zh"
    CHINESE_CHINA = "zh-CN"
    CHINESE_HONG_KONG = "zh-HK"
    CHINESE_SIMPLIFIED = "zh-Hans"
    CHINESE_SINGAPORE = "zh-SG"
    CHINESE_TAIWAN = "zh-TW"
    CHINESE_TRADITIONAL = "zh-Hant"
    CHOCTAW = "cho"
    COPTIC = "cop"
    CORSICAN = "co"
    CREE = "cr"
    CROATIAN = "hr"
    CZECH = "cs"
    DANISH = "da"
    DOGRI = "doi"
    DUTCH = "nl"
    DUTCH_BELGIUM = "nl-BE"
    DUTCH_NETHERLANDS = "nl-NL"
    DZONGKHA = "dz"
    ENGLISH = "en"
    ENGLISH_AUSTRALIA = "en-AU"
    ENGLISH_CANADA = "en-CA"
    ENGLISH_INDIA = "en-IN"
    ENGLISH_IRELAND = "en-IE"
    ENGLISH_UNITED_KINGDOM = "en-GB"
    ENGLISH_UNITED_STATES = "en-US"
    ESPERANTO = "eo"
    ESTONIAN = "et"
    EWE = "ee"
    FAROESE = "fo"
    FIJIAN = "fj"
    FILIPINO = "fil"
    FINNISH = "fi"
    FRENCH = "fr"
    FRENCH_BELGIUM = "fr-BE"
    FRENCH_CANADA = "fr-CA"
    FRENCH_FRANCE = "fr-FR"
    FRENCH_SWITZERLAND = "fr-CH"
    FULA = "ff"
    GALICIAN = "gl"
    GANDA = "lg"
    GEORGIAN = "ka"
    GERMAN = "de"
    GERMAN_AUSTRIA = "de-AT"
    GERMAN_GERMANY = "de-DE"
    GERMAN_SWITZERLAND = "de-CH"
    GREEK = "el"
    GUARANI = "gn"
    GUJARATI = "gu"
    GUSII = "guz"
    HAITIAN_CREOLE = "ht"
    HAKKA_CHINESE = "hak"
    HAKKA_CHINESE_TAIWAN = "hak-TW"
    HARYANVI = "bgc"
    HAUSA = "ha"
    HAWAIIAN = "haw"
    HEBREW = "he"
    HINDI = "hi"
    HINDI_LATIN = "hi-Latn"
    HIRI_MOTU = "ho"
    HUNGARIAN = "hu"
    ICELANDIC = "is"
    IGBO = "ig"
    INDONESIAN = "id"
    INTERLINGUA = "ia"
    INTERLINGUE = "ie"
    INUKTITUT = "iu"
    INUPIAQ = "ik"
    IRISH = "ga"
    ITALIAN = "it"
    JAPANESE = "ja"
    JAVANESE = "jv"
    KALAALLISUT = "kl"
    KALENJIN = "kln"
    KAMBA = "kam"
    KANNADA = "kn"
    KASHMIRI = "ks"
    KAZAKH = "kk"
    KHMER = "km"
    KIKUYU = "ki"
    KINYARWANDA = "rw"
    KLINGON = "tlh"
    KONKANI = "kok"
    KOREAN = "ko"
    KURDISH = "ku"
    KYRGYZ = "ky"
    LADINO = "lad"
    LAO = "lo"
    LATIN = "la"
    LATVIAN = "lv"
    LINGALA = "ln"
    LITHUANIAN = "lt"
    LUBA_KATANGA = "lu"
    LUO = "luo"
    LUXEMBOURGISH = "lb"
    LUYIA = "luy"
    MACEDONIAN = "mk"
    MAITHILI = "mai"
    MALAGASY = "mg"
    MALAY = "ms"
    MALAY_SINGAPORE = "ms-SG"
    MALAYALAM = "ml"
    MALTESE = "mt"
    MANIPURI = "mni"
    MAORI = "mi"
    MARATHI = "mr"
    MASAI = "mas"
    MERU = "mer"
    MIN_NAN_CHINESE = "nan"
    MIN_NAN_CHINESE_TAIWAN = "nan-TW"
    MIXE = "mco"
    MIZO = "lus"
    MONGOLIAN = "mn"
    MONGOLIAN_MONGOLIAN = "mn-M"
    NAURU = "na"
    NAVAJO = "nv"
    NEPALI = "ne"
    NIGERIAN_PIDGIN = "pcm"
    NORTH_NDEBELE = "nd"
    NORTHERN_SOTHO = "nso"
    NORWEGIAN = "no"
    OCCITAN = "oc"
    ODIA = "or"
    OROMO = "om"
    PAPIAMENTO = "pap"
    PASHTO = "ps"
    PERSIAN = "fa"
    PERSIAN_AFGHANISTAN = "fa-AF"
    PERSIAN_IRAN = "fa-IR"
    POLISH = "pl"
    PORTUGUESE = "pt"
    PORTUGUESE_BRAZIL = "pt-BR"
    PORTUGUESE_PORTUGAL = "pt-PT"
    PUNJABI = "pa"
    QUECHUA = "qu"
    ROMANIAN = "ro"
    ROMANIAN_MOLDOVA = "ro-MD"
    ROMANSH = "rm"
    RUNDI = "rn"
    RUSSIAN = "ru"
    RUSSIAN_LATIN = "ru-Latn"
    SAMOAN = "sm"
    SANGO = "sg"
    SANSKRIT = "sa"
    SANTALI = "sat"
    SARDINIAN = "sc"
    SCOTTISH_GAELIC = "gd"
    SERBIAN = "sr"
    SERBIAN_CYRILLIC = "hr-Cyrl"
    SERBIAN_LATIN = "sr-Latn"
    SERBO_CROATIAN = "sh"
    SHERDUKPEN = "sdp"
    SHONA = "sn"
    SICILIAN = "scn"
    SINDHI = "sd"
    SINHALA = "si"
    SLOVAK = "sk"
    SLOVENIAN = "sl"
    SOMALI = "so"
    SOUTH_NDEBELE = "nr"
    SOUTHERN_SOTHO = "st"
    SPANISH = "es"
    SPANISH_LATIN_AMERICA = "es-419"
    SPANISH_MEXICO = "es-MX"
    SPANISH_SPAIN = "es-ES"
    SPANISH_UNITED_STATES = "es-US"
    SUNDANESE = "su"
    SWAHILI = "sw"
    SWATI = "ss"
    SWEDISH = "sv"
    TAGALOG = "tl"
    TAJIK = "tg"
    TAMIL = "ta"
    TATAR = "tt"
    TELUGU = "te"
    THAI = "th"
    TIBETAN = "bo"
    TIGRINYA = "ti"
    TOK_PISIN = "tpi"
    TOKI_PONA = "tok"
    TONGAN = "to"
    TSONGA = "ts"
    TSWANA = "tn"
    TURKISH = "tr"
    TURKMEN = "tk"
    TWI = "tw"
    UKRAINIAN = "uk"
    URDU = "ur"
    UYGHUR = "ug"
    UZBEK = "uz"
    VENDA = "ve"
    VIETNAMESE = "vi"
    VOLAPÜK = "vo"
    VÕRO = "vro"
    WELSH = "cy"
    WESTERN_FRISIAN = "fy"
    WOLAYTTA = "wal"
    WOLOF = "wo"
    XHOSA = "xh"
    YIDDISH = "yi"
    YORUBA = "yo"
    ZULU = "zu"


class LicenseEnum(str, Enum):
    STANDARD = "standard"
    CREATIVE_COMMONS = "creative_commons"


class AllowCommentsEnum(str, Enum):
    ALL_COMMENTS = "ALL_COMMENTS"
    HOLD_INAPPROPRIATE = "AUTOMATED_COMMENTS"
    HOLD_INAPPROPRIATE_STRICT = "AUTO_MODERATED_COMMENTS_HOLD_MORE"
    HOLD_ALL = "APPROVED_COMMENTS"


class CommentsSortOrderEnum(str, Enum):
    LATEST = "MDE_COMMENT_SORT_ORDER_LATEST"
    TOP = "MDE_COMMENT_SORT_ORDER_TOP"


class ThumbnailFormatEnum(str, Enum):
    PNG = "CUSTOM_THUMBNAIL_IMAGE_FORMAT_PNG"
    JPG = "CUSTOM_THUMBNAIL_IMAGE_FORMAT_JPEG"


class PremiereDurationEnum(str, Enum):
    ONE_MIN = "60"
    TWO_MIN = "120"
    THREE_MIN = "180"
    FOUR_MIN = "240"
    FIVE_MIN = "300"


class PremiereThemeEnum(str, Enum):
    CLASSIC = "VIDEO_PREMIERE_INTRO_THEME_DEFAULT"
    ALTERNATIVE = "VIDEO_PREMIERE_INTRO_THEME_ALTERNATIVE"
    AMBIENT = "VIDEO_PREMIERE_INTRO_THEME_AMBIENT"
    BRIGHT = "VIDEO_PREMIERE_INTRO_THEME_BRIGHT"
    CALM = "VIDEO_PREMIERE_INTRO_THEME_CALM"
    CINEMATIC = "VIDEO_PREMIERE_INTRO_THEME_CINEMATIC"
    CONTEMPORARY = "VIDEO_PREMIERE_INTRO_THEME_CONTEMPORARY"
    DRAMATIC = "VIDEO_PREMIERE_INTRO_THEME_DRAMATIC"
    FUNKY = "VIDEO_PREMIERE_INTRO_THEME_FUNKY"
    GENTLE = "VIDEO_PREMIERE_INTRO_THEME_GENTLE"
    HAPPY = "VIDEO_PREMIERE_INTRO_THEME_HAPPY"
    INSPIRATIONAL = "VIDEO_PREMIERE_INTRO_THEME_INSPIRATIONAL"
    KIDS = "VIDEO_PREMIERE_INTRO_THEME_KIDS"
    SCI_FI = "VIDEO_PREMIERE_INTRO_THEME_SCI_FI"
    SPORTS = "VIDEO_PREMIERE_INTRO_THEME_SPORTS"


enum_classes = [
    PrivacyEnum,
    CategoryEnum,
    LanguageEnum,
    LicenseEnum,
    AllowCommentsEnum,
    CommentsSortOrderEnum,
    ThumbnailFormatEnum,
    PremiereDurationEnum,
    PremiereThemeEnum,
]


def enum_allow_key__new__(cls, value):
    try:
        return Enum.__new__(cls, value)
    except ValueError:
        # maybe key was specified instead of value?
        try:
            value = getattr(cls, value)
            return Enum.__new__(cls, value)
        except (AttributeError, TypeError):
            raise ValueError(f"{cls.__name__} has no key or value '{value}'")


for enum_class in enum_classes:
    setattr(enum_class, "__new__", enum_allow_key__new__)


@dataclass_json
@dataclass
class Playlist:
    """Metadata of playlist to create and/or add video to"""

    title: str
    """Title. Max length 150"""

    description: str = ""
    """Description. Max length 5000"""

    privacy: PrivacyEnum = PrivacyEnum.PUBLIC
    """Privacy. Possible values: PUBLIC, UNLISTED, PRIVATE"""

    create_if_title_exists: bool = False
    """Whether to create playlist if a playlist with the same
    title already exists on the channel"""

    create_if_title_doesnt_exist: bool = True
    """Whether to create playlist if there is no playlist with the same title"""


@dataclass_json
@dataclass
class CaptionsFile:
    """Subtitles file"""

    path: str
    """Path to .srt file"""

    language: Optional[LanguageEnum] = None
    """Language of captions. If None, language will default to audio language"""


@dataclass_json
@dataclass
class Metadata:
    """Metadata of video to upload"""

    title: str
    """Title. Max length 100. Cannot contain < or > characters"""

    description: str = ""
    """Description. Max length 5000. Cannot contain < or > characters"""

    privacy: PrivacyEnum = PrivacyEnum.PRIVATE
    """Privacy. Possible values: PUBLIC, UNLISTED, PRIVATE"""

    made_for_kids: bool = False
    """Made for kids. If true comments will be disabled"""

    tags: List[str] = ()
    """List of tags"""

    # optional metadata for update_metadata
    scheduled_upload: Optional[datetime.datetime] = field(
        default=None,
        metadata=config(
            decoder=lambda x: (
                datetime.datetime.fromisoformat(x) if x is not None else None
            ),
            encoder=lambda x: datetime.datetime.isoformat(x) if x is not None else None,
            mm_field=mm_field.DateTime("iso", allow_none=True),
        ),
    )
    """
    Date to make upload public. If set, video will be set to private until the date, unless video
    is a premiere in which case it will be set to public. Video will not be a premiere unless both
    premiere_countdown_duration and premiere_theme are set
    """

    premiere_countdown_duration: Optional[PremiereDurationEnum] = None
    """Duration of premiere countdown in seconds. Possible values: 60, 120, 180, 240, 300"""

    premiere_theme: Optional[PremiereThemeEnum] = None
    """
    Theme of premiere countdown. Possible values: CLASSIC, ALTERNATIVE, AMBIENT, BRIGHT, CALM,
    CINEMATIC, CONTEMPORARY, DRAMATIC, FUNKY, GENTLE, HAPPY, INSPIRATIONAL, KIDS, SCI_FI, SPORTS
    """

    playlist_ids: Optional[List[str]] = None
    """List of existing playlist IDs to add video to"""

    playlists: Optional[List[Playlist]] = None
    """List of playlists to create and/or add video to"""

    thumbnail: Optional[str] = None
    """Path to thumbnail file to upload"""

    publish_to_feed: Optional[bool] = None
    """Whether to notify subscribers"""

    category: Optional[CategoryEnum] = None
    """Category. Category-specific metadata is not supported yet. Possible values: FILM_ANIMATION,
    AUTOS_VEHICLES, MUSIC, PETS_ANIMALS, SPORTS, TRAVEL_EVENTS, GAMING, PEOPLE_BLOGS, COMEDY,
    ENTERTAINMENT, NEWS_POLITICS, HOWTO_STYLE, EDUCATION, SCIENCE_TECH, NONPROFITS_ACTIVISM"""

    auto_chapter: Optional[bool] = None
    """Whether to use automatic video chapters"""

    auto_places: Optional[bool] = None
    """Whether to use automatic places"""

    auto_concepts: Optional[bool] = None
    """Whether to use automatic concepts"""

    has_product_placement: Optional[bool] = None
    """Whether video has product placement"""

    show_product_placement_overlay: Optional[bool] = None
    """Whether to show product placement overlay"""

    recorded_date: Optional[datetime.date] = field(
        default=None,
        metadata=config(
            decoder=lambda x: datetime.date.fromisoformat(x) if x is not None else None,
            encoder=lambda x: datetime.date.isoformat(x) if x is not None else None,
            mm_field=mm_field.Date("iso", allow_none=True),
        ),
    )
    """Day, month, and year that video was recorded"""

    restricted_to_over_18: Optional[bool] = None
    """Whether video is age restricted"""

    audio_language: Optional[LanguageEnum] = None
    """Language of audio"""

    captions_files: Optional[List[CaptionsFile]] = None
    """Path to captions files (.srt)"""

    license: Optional[LicenseEnum] = None
    """License. Possible values: STANDARD, CREATIVE_COMMONS"""

    allow_comments: Optional[bool] = None
    """Whether to allow comments"""

    allow_comments_mode: Optional[AllowCommentsEnum] = None
    """Comment filtering mode. Possible values: ALL_COMMENTS, HOLD_INAPPROPRIATE,
    HOLD_INAPPROPRIATE_STRICT, HOLD_ALL"""

    can_view_ratings: Optional[bool] = None
    """Whether video likes/dislikes can be seen"""

    comments_sort_order: Optional[CommentsSortOrderEnum] = None
    """Default comment sort order. Possible values: LATEST, TOP"""

    allow_embedding: Optional[bool] = None
    """Whether to allow embedding on 3rd party sites"""

    def validate(self):
        """Raises error if metadata is invalid"""
        if (
            self.premiere_countdown_duration is not None
            or self.premiere_theme is not None
        ):
            if None in (
                self.premiere_countdown_duration,
                self.premiere_theme,
                self.scheduled_upload,
            ):
                raise ValueError(
                    "If trying to upload a premiere, premiere_countdown_duration, "
                    "premiere_theme, and scheduled_upload must be set"
                )

        if self.captions_files is not None:
            for caption_file in self.captions_files:
                if caption_file.language is None and self.audio_language is None:
                    raise ValueError(
                        "Must either specify captions file language or audio_language"
                    )

        if self.restricted_to_over_18 and self.made_for_kids:
            raise ValueError(
                "Video cannot be made for kids and also restricted to over 18"
            )

        if len(self.title) > 100:
            raise ValueError("Title must be at most 100 characters long")

        if len(self.description) > 5000:
            raise ValueError("Description must be at most 5000 characters long")

        if any(c in s for c in "<>" for s in (self.title, self.description)):
            raise ValueError("Title and description cannot contain angled brackets")

        errors = self.schema().validate(self.to_dict())
        if errors:
            raise ValueError(f"{errors}")


__all__ = [
    "Metadata",
    "Playlist",
    "CaptionsFile",
    "PremiereThemeEnum",
    "PremiereDurationEnum",
    "ThumbnailFormatEnum",
    "CommentsSortOrderEnum",
    "AllowCommentsEnum",
    "LicenseEnum",
    "LanguageEnum",
    "CategoryEnum",
    "PrivacyEnum",
]
